/*  $Id: COut.cc,v 1.32 2000/01/07 03:37:00 shaggy Exp $
 *      C++ output
 *      
 *      relative path: lib/cparse
 *  
 *  Copyright (c) 1996 Frank E. Barrus.  All rights reserved.
 *  
 *  This file is part of the ShagOS distribution, and may be freely used and
 *  distributed by the terms of the ShagWare General Public License (SGPL),
 *  version 2, or, at your option, any newer official version of that License.
 *  
 *  THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT WARRANTY, EXPRESS OR IMPLIED.
 *  THE AUTHOR SPECIFICALLY DISCLAIMS ANY IMPLIED WARRANTIES OF ANY KIND,
 *  INCLUDING BUT NOT LIMITED TO THE IMPLIED WARRANTIES OF MERCHANTABILITY AND
 *  FITNESS FOR A PARTICULAR PURPOSE, AND WILL NOT BE LIABLE FOR ANY DAMAGES.
 *  For more warranty details, see the ShagWare General Public License, which
 *  should have been received with this program, in a file called 'lic_spl2.txt'
 *  If this license is missing, please contact the author and describe how
 *  this software was obtained.  Email: shaggy@csh.rit.edu
 *  
 *  Complete, unaltered, original distributions of this and related code
 *  may be obtained from:   http://www.csh.rit.edu/~shaggy/software.html
 *  or:	ftp://ftp.csh.rit.edu/pub/csh/shaggy
 */

#include <libc.h>
#include <stdx1.h>
#include <defines>
#include <ldefines>

#include <CObj>
#include <CLex>
#include <CParse>
#include <vCAction>
#include <TokenList>	// perhaps should be part of CAction
#include <StrTab>
#include <SymTab>
#include <assert.h>

#include <CRep>

#include "COut.hh"


#define INTERNAL_ERROR	{ _err.internal_error(__FILE__, __LINE__, __FUNCTION__); \
			  for(;;) ; }

// lexical shortcuts
typedef vStrTab::Id Id;
typedef vCAction::Token Token;
typedef CLex::Token LToken;
typedef vCAction::Item Item;


sz *
strpcat(sz *str, csz *prefix)
{
	size_t n = strlen(prefix);
	memmove(str+n, str, strlen(str)+1);
	memcpy(str, prefix, n);
	return str;
}


struct objtn { obj_Integral *o; csz *n; } itypes[] = {
	{&Type_char, "char"},
	{&Type_schar, "signed char"},
	{&Type_uchar, "unsigned char"},
	{&Type_int, "int"},
	{&Type_uint, "unsigned int"},
	{&Type_short, "short"},
	{&Type_ushort, "unsigned short"},
	{&Type_long, "long"},
	{&Type_ulong, "unsigned long"},
	{&Type_long2, "long long"},
	{&Type_ulong2, "unsigned long long"},
	{0, 0},
};


csz*
simple_typename(type_t type)
{
	for(objtn *ip = itypes; ip->n; ip++) {
		if(ip->o == type)
			return ip->n;
	}
	return "?";
}

void
CCOut::_assertfail(csz *cond, csz *file, int line)
{
	sz buf[MAXIDLEN];

	onl();
	_err.error("Assertion \"%s\" failed (%s:%d)", cond, file, line);
	if(_curdecl)
		_err.error("Failed for %s", scoped_string(_curdecl, buf, true));
	_err.internal_error(file, line, 0);
}

sz *
CCOut::literal_string(type_t type, uint64 val, sz *s)
{
	sz *saveobuf = _obuf;
	_obuf = s;
	printliteral(type, val);
	_obuf = saveobuf;
	return s;
}


sz *
CCOut::dbg_type_string(type_t type, sz *s)
{
	assert(s);

	if(!type)
		strcat(s, "<null>");
	else if(Method *meth = dcast_Method(type)) {
		strcat(s, "Method(");
		s = dbg_type_string(meth->otype(), s);
		strcat(s, ",");
		s = dbg_type_string(meth->ftype(), s);
		strcat(s, ")");
	} else if(Func *func = dcast_Func(type)) {
		strcat(s, "Func(");
		if(func->scope())
			scoped_string(func->scope(), s+strlen(s));
		else
			strcat(s, "<null>");
		strcat(s, ", ");
		dbg_type_string(func->rtype(), s);
		strcat(s, ", ");
		dbg_type_string(func->psig(), s);
	} else if(Params *par = dcast_Params(type)) {
		strcat(s, "Params(");
		for(uint i=0; i<par->nparam(); i++) {
			if(i)
				strcat(s, ", ");
			s = dbg_type_string(par->ptype(i), s);
		}
		if(par->variable())
			strcat(s, "...");
		strcat(s, ")");
	} else if(Array *array = dcast_Array(type)) {
		sprintf(s+strlen(s), "Array(%d, ", array->size());
		s = dbg_type_string(array->rtype(), s);
		strcat(s, ")");
	} else if(PtrTo *ptr = dcast_PtrTo(type)) {
		strcat(s, "PtrTo(");
		if(ptr->qvolatile()) strcat(s, "v, ");
		if(ptr->qconst()) strcat(s, "c, ");
		s = dbg_type_string(ptr->rtype(), s);
		strcat(s, ")");
	} else if(RefTo *ref = dcast_RefTo(type)) {
		strcat(s, "RefTo(");
		if(ref->qvolatile()) strcat(s, "v, ");
		if(ref->qconst()) strcat(s, "c, ");
		s = dbg_type_string(ref->rtype(), s);
		strcat(s, ")");
	} else if(Qualifier *qual = dcast_Qualifier(type)) {
		strcat(s, "Qual(");
		if(qual->qvolatile()) strcat(s, "v, ");
		if(qual->qconst()) strcat(s, "c, ");
		s = dbg_type_string(qual->rtype(), s);
		strcat(s, ")");
	} else if(DeclRef *dref = dcast_DeclRef(type)) {
		strcat(s, "DeclRef(");
		scoped_string(dref->decl(), s+strlen(s));
		strcat(s, ")");
	} else if(Aggr *aggr = dcast_Aggr(type)) {
		strcat(s, "Aggr(");
		if(aggr->name())
			strcat(s, cstr(aggr->name()));
		else
			strcat(s, "<null>");
		if(aggr->scope())
			sprintf(s+strlen(s), "%p", aggr->scope());
		strcat(s, ")");
	} else if(Enum *enm = dcast_Enum(type)) {
		strcat(s, "Enum(");
		scoped_string(enm->decl(), s+strlen(s));
		strcat(s, ")");
	} else if(Integral *integ = dcast_Integral(type)) {
		sprintf(s+strlen(s), "Integral(%d,%d,%c,%c)",
			integ->bitsize(), integ->alignment(),
			integ->sign() ? 's' : 'u',
			integ->bigendian() ? 'b' : 'l');
	} else if(type == &Type_bool)
		strcat(s, "bool");
	else if(type == &Type_wchar)
		strcat(s, "wchar_t");
	else if(type == &Type_float)
		strcat(s, "float");
	else if(type == &Type_double)
		strcat(s, "double");
	else if(type == &Type_void)
		strcat(s, "void");
	else if(type->isnull())
			strcat(s, "<null type>"); 
	else {
		sprintf(s+strlen(s), "<unknown:%p/%#llx>",
				type->type(), type->data());
	}
	return s;
}

sz *
CCOut::type_string(type_t type, sz *s)
{
	assert(s);

	sz buf[MAXTYPELEN];

	if(_debugtypes) {
		if(*s)
			strcat(s, ": ");
		return dbg_type_string(type, s);
	}
	if(!type)
		return s;
	if(Method *meth = dcast_Method(type)) {
		type_t otype = meth->otype();
		bool qconst = false;
		bool qvolatile = false;
		while(Qualifier *qual = dcast_Qualifier(otype)) {
			if(qual->qconst())
				qconst = true;
			if(qual->qvolatile())
				qvolatile = true;
			otype = qual->rtype();
		}
		if(s[0] == '*') {
			type_t ot = meth->otype();
			while(Qualifier *qual = dcast_Qualifier(ot))
				ot = qual->rtype();
			strpcat(s, "::");
			s = type_string(ot, s);
			strpcat(s, "(");
			strcat(s, ")");
		}
		s = type_string(meth->ftype(), s);
		if(qconst)
			strcat(s, " const");
		if(qvolatile)
			strcat(s, " volatile");
	} else if(Func *func = dcast_Func(type)) {
		if(s[0] == '*' || s[0] == '&') {
			strpcat(s, "(");
			strcat(s, ")");
		}
		Scope savescope = scope();
		assert(func->scope());
		scope(func->scope());
		strcat(s, "(");
		for(uint i=0; i<func->nparam(); i++) {
			if(i)
				strcat(s, ", ");
			Decl *d = param_decl(func, i);
			if(d && d->id())
				strcpy(buf, cstr(d->id()));
			else 
				buf[0] = '\0';
			strcat(s, type_string(func->ptype(i),
						buf));
			if(d && d->initialized()) {
				strcat(s, " = ");
				if(d->storage() == TDecl::sc_constant)
					sprintf(s+strlen(s), "%lld", d->cval());
				else {
					sz *saveobuf = _obuf;	
					_obuf = s+strlen(s);
					if(_debugexpr)
						Expr(d->vexpr()).dump();
					printexpr(d->vexpr());
					_obuf = saveobuf;
				}
			}
		}
		if(func->variable()) {
			if(func->nparam())
				strcat(s, ", ");
			strcat(s, "...");
		}
		strcat(s, ")");
		scope(savescope);
		s = type_string(func->rtype(), s);
	} else if(Params *par = dcast_Params(type)) {
		if(s[0] == '*' || s[0] == '&') {
			strpcat(s, "(");
			strcat(s, ")");
		}
		strcat(s, "(");
		for(uint i=0; i<par->nparam(); i++) {
			if(i)
				strcat(s, ", ");
			buf[0] = '\0';
			strcat(s, type_string(par->ptype(i),
						buf));
		}
		if(par->variable()) {
			if(par->nparam())
				strcat(s, ", ");
			strcat(s, "...");
		}
		strcat(s, ")");
	} else if(Array *array = dcast_Array(type)) {
		if(s[0] == '*' || s[0] == '&') {
			strpcat(s, "(");
			strcat(s, ")");
		}
		if(array->size() == NPOS) {
			Expr e = reinterpret_cast<ExprNode*>(array->dsize());
			if(e) {
				buf[0] = '[';
				buf[1] = '\0';
				sz *saveobuf = _obuf;
				_obuf = &buf[1];
				if(_debugexpr)
					e.dump();
				printexpr(e);
				_obuf = saveobuf;
				strcat(buf, "]");
			} else
				sprintf(buf, "[]");
		} else if(0 == array->lower()) 
			sprintf(buf, "[%d]", array->size());
		else
			sprintf(buf, "[%d..%d]", array->lower(),
					array->upper());
		strcat(s, buf);
		s = type_string(array->rtype(), s);
	} else if(PtrTo *ptr = dcast_PtrTo(type)) {
		if(ptr->qvolatile()) strpcat(s, "volatile ");
		if(ptr->qconst()) strpcat(s, "const ");
		strpcat(s, "*");
		s = type_string(ptr->rtype(), s);
	} else if(RefTo *ref = dcast_RefTo(type)) {
		if(ref->qvolatile()) strpcat(s, "volatile ");
		if(ref->qconst()) strpcat(s, "const ");
		strpcat(s, "&");
		s = type_string(ref->rtype(), s);
	} else if(Qualifier *qual = dcast_Qualifier(type)) {
		type_t rtype = qual->rtype();
		s = type_string(rtype, s);
		if(rtype->istype(Type_Func)) {
			if(qual->qvolatile()) strcat(s, " volatile");
			if(qual->qconst()) strcat(s, " const");
		} else {
			if(qual->qvolatile()) strpcat(s, "volatile ");
			if(qual->qconst()) strpcat(s, "const ");
		}
	} else {
		if(*s) strpcat(s, " ");
		if(DeclRef *dref = dcast_DeclRef(type)) {
			sz buf[MAXIDLEN];
			Decl *decl = dref->decl();
			strpcat(s, scoped_string(decl, buf));
		} else if(Aggr *aggr = dcast_Aggr(type)) {
			/* XXX - needs scope */
			if(aggr->name())
				strpcat(s, cstr(aggr->name()));
			else if (aggr->scope()) {
				Decl *d = Scope(aggr->scope()).decl();
				strpcat(s, cstr(d->id()));
			} else {	
				sprintf(buf, "anonymous%lx", (uintptr)aggr->scope());
				strpcat(s, buf);
			}
			if(aggr->scope()) {
				sz buf[MAXIDLEN];
				scoped_string(Scope(aggr->scope()).parent(), buf);
				strpcat(s, buf);
			}
		} else if(Enum *enm = dcast_Enum(type)) {
			sz buf[MAXIDLEN];
			strpcat(s, scoped_string(enm->decl(), buf));
		} else if(type->istype(Type_Integral)) 
			strpcat(s, simple_typename(type));
		else if(type == &Type_bool)
			strpcat(s, "bool");
		else if(type == &Type_wchar)
			strpcat(s, "wchar_t");
		else if(type == &Type_float)
			strpcat(s, "float");
		else if(type == &Type_double)
			strpcat(s, "double");
		else if(type == &Type_void)
			strpcat(s, "void");
		/* else if(type->isnull())
			strpcat(s, "<null>");  */
		else {
			sprintf(buf, "<unknown:%p/%#llx>",
				type->type(), type->data());
			strpcat(s, buf);
		}
	}
	return s;
}

/*
sz*
CCOut::fullscoped_string(Id id, Scope sc, sz *s)
{
	Scope parent = sc.parent();
	if(parent) {
		s = scoped_string(sc.pchid(), parent, s);
		strcat(s, "::");
	} else
		strcpy(s, "::");
	strcat(s, cstr(id));
	return s;
}
*/


sz*
CCOut::scoped_string(Scope sc, sz *s)
{
	assert(sc);
	assert(s);

	Scope parent = sc.parent();
	if(sc == scope()) 
		*s = '\0';
	else if(!parent) 
		strcpy(s, "::");
	else {
		Decl *d = sc.decl();
		Scope st = 0;
		csz *sn;
		if(!d)
			sn = "{";
		else if(!d->id())
			sn = "?";
		else if(d->id() == ID_ANON)
			sn = "<anon>";
		else if(d->id() == ID_FRIEND)
			sn = "<friend>";
		else {
			sn = cstr(d->id());
			if(Decl *nd = scope().tfind(d->id())) {
				type_t t = nd->ptype();
				if(Aggr *aggr = dcast_Aggr(t)) 
					st = aggr->scope();
			}
		}
		sz *sp = s;
		if(st != sc || _fullscope) 
			sp += strlen(scoped_string(parent, sp));
		strcpy(sp, sn);
		if(d && d->isfunc() && !_debugtypes)
			type_string(d->rtype(), sp);
		strcat(sp, "::");
	}
	return s;
}


static bool
canfind(Decl *decl, const Scope inscope, bool withtag)
{
	Decl *d;
	d = inscope.tfind(decl->id(), withtag);
	while(d && d != decl && !d->odeclared() && !inscope.pchid())
		d = d->link();
	if(!d && inscope.parent())
		return canfind(decl, inscope.parent(), withtag);
	return d == decl || (d && d->scope() == decl->scope());
}

sz*
CCOut::scoped_string(Decl *decl, sz *s, bool needtag)
{
	assert(decl);
	assert(s);

	Scope sc = decl->scope();
	if(_fullscope) {
		needtag = true;
		scoped_string(sc, s);
	} else if(canfind(decl, scope(), false)) {
		*s = '\0';
	} else {
		if(decl->tagged() && canfind(decl, scope(), true)) {
			*s = '\0';
			needtag = true;
		} else {
			if(decl->tagged() && !canfind(decl, sc, false))
				needtag = true;
			scoped_string(sc, s);
		}
	}
	{
		csz *ss = cstr(decl->id());
		while(*ss == ' ')
			++ss;
		strcat(s, ss);
	}
	if(needtag && decl->tagged()) {
		strpcat(s, " ");
		strpcat(s, decl->dtname());
	} else {
		switch(decl->dtype()) {
		case Decl::dt_type: strpcat(s, "/*type*/ "); break;
		case Decl::dt_dclass: strpcat(s, "/*dclass*/ "); break;
		default: ;
		}
	}
	return s;
}


void
CCOut::onl(int y)
{
	while(_nl < y) {
		if(_obuf) 
			strcat(_obuf, " ");
		else
			printf("\n");
		_x = 0;
		++_nl;
	}
	_ws = false;
}

void
CCOut::onltab(int x)
{
	onl();
	if(x < 0)
		x = _indent;
	if(x < _x) {
		if(!_obuf)
			printf("\n");
		++_nl;
		_x = 0;
	}
	if(x > _x && !_obuf)
		printf("%*s", x-_x, "");
	_x = x;
}


int
CCOut::oprintf(csz *fmt, ...)
{
	sz buf[MAXLINELEN];
	sz *b = buf;
	int n;
	if(_sectionlabel && !_obuf) {
		onl(2);
		csz *s = _sectionlabel;
		_sectionlabel = 0;
		oprintf("// %s", s);
		onl();
	}
	if(!_obuf && _x == 0)
		onltab();
	if(_ws) {
		_ws = false;
		if(!_nl)
			oprintf(" ");
	}
	if(_obuf) 
		b = _obuf+strlen(_obuf);
	va_list ap;
	va_start(ap, fmt);
	n = vsprintf(b, fmt, ap);
	va_end(ap);
	if(n > 0 && !_obuf) {
		/* better splitting logic could go here */
		if(_wid && _x > _indent+1 && _x + n >= _wid) {
			onltab();
			oprintf(" ");
		}
		printf("%s", buf);
		_nl = 0;
		_x += n;
	}
	return n;
}


void
CCOut::printliteral(type_t type, uint64 val)
{
	assert(type);

	while(Qualifier *qual = dcast_Qualifier(type)) 
		type = qual->rtype();
	csz *fmt;
	bool use64 = false;
	if(type == &Type_int) 
		fmt = "%d";
	else if(type == &Type_uint)
		fmt = "%uU";
	else if(type == &Type_long)
		fmt = "%ldL";
	else if(type == &Type_ulong)
		fmt = "%luUL";
	else if(type == &Type_long2) {
		fmt = "%lldLL"; use64 = true;
	} else if(type == &Type_ulong2) {
		fmt = "%lluULL"; use64 = true;
	} else if(type == &Type_char) {
		char ch = val;
		oprintf("%s", _clex.token_string(CLex::CHAR, &ch, 1));
		return;
	} else if(type == &Type_bool) {
		oprintf(val ? "true" : "false");
		return;
	} else if(type == &Type_Str) {
		oprintf("%s", _clex.token_string(CLex::STR, cstr((vStrTab::Id)val)));
		return;
	} else {
		sz buf[MAXTYPELEN];
		buf[0] = '\0';
		if((int64)val >= -1 && (int64)val <= 1)
			fmt = "((%s)%lld)";
		else
			fmt = "((%s)0x%llx)";
		oprintf(fmt, type_string(type, buf), val);
		return;
	}
	if(use64)
		oprintf(fmt, val);
	else
		oprintf(fmt, (uint32)val);
}


void
CCOut::printexprlist(ExprList exprlist)
{
	bool line = exprlist.fmt() == T(LINE);
	if(line) {
		rshift();
		onl();
	}
	for(Expr e = exprlist.first(); e; e++) {
		printexpr(e);
		if(e.next()) {
			oprintf(",");
			ws();
		}
		if(line)
			onl();
	}
	if(line)
		lshift();
}

void
CCOut::printstmtlist(StmtList stmtlist)
{
	const Scope savescope = scope();
	if(stmtlist.scope())
		scope(stmtlist.scope());
	oprintf("{");
	rshift();
	for(Stmt s = stmtlist.first(); s; s++) {
		onl();
		printstmt(s);
	}
	lshift();
	onl();
	oprintf("}");
	scope(savescope);
}

void
CCOut::printdecllist(StmtList stmtlist)
{
	const Scope savescope = scope();
	bool first = true;
	if(stmtlist.scope())
		scope(stmtlist.scope());
	for(Stmt s = stmtlist.first(); s; s++) {
		assert(s.op() == T(DECL));
		Decl *d = s.left().decl();
		if(first) 
			define_sym(d, false);
		else {
			oprintf(", %s = ", cstr(d->id()));
			printexpr(d->vexpr());	/* XXX - const? */
		}	
		d->odeclared(true);
		d->odefined(true);
		first = false;
	}
	scope(savescope);
}

void
CCOut::printctorinit(StmtList stmtlist)
{
	if(stmtlist.empty())
		return;
	const Scope savescope = scope();
	if(stmtlist.scope())
		scope(stmtlist.scope());
	rshift();
	onltab(_indent-2);
	oprintf(": ");
	for(Stmt s = stmtlist.first(); s; ) {
		sz buf[MAXIDLEN];
		assert(s.left().op() == T(VAL_ID));
		oprintf("%s", scoped_string(s.left().decl(), buf, false));
		oprintf("(");
		printexpr(s.right());
		oprintf(")");
		if(++s) {
			oprintf(",");
			onl();
		}
	}
	lshift();
	onl();
	scope(savescope);
}

void
CCOut::printstmt(Stmt stmt)
{
	if(!stmt)
		return;
	switch(stmt.op()) {
	case T(KW_IF):
		oprintf("if");
		printexpr(stmt.left(), true);
		printstmti(Stmt(stmt.mid()));
		if(stmt.right()) {
			if(stmt.mid().op() == T(STMTLIST))
				oprintf(" ");
			else
				onl();
			oprintf("else");
			if(stmt.right().op() == T(KW_IF)) {
				oprintf(" ");
				printstmt(Stmt(stmt.right()));
			} else
				printstmti(Stmt(stmt.right()));
		}
		break;
	case T(KW_DO):
		oprintf("do");
		if(stmt.mid().op() != T(STMTLIST)) {
			oprintf(" ");
			printstmt(Stmt(stmt.mid()));
		} else
			printstmti(Stmt(stmt.mid()));
		oprintf(" while");
		printexpr(stmt.right(), true);
		oprintf(";");
		break;
	case T(KW_WHILE):
		oprintf("while");
		printexpr(stmt.left(), true);
		printstmti(Stmt(stmt.mid()));
		break;
	case T(KW_SWITCH):
		oprintf("switch");
		printexpr(stmt.left(), true);
		printstmti(Stmt(stmt.mid()));
		break;
	case T(KW_FOR): {
		const Scope savescope = scope();
		Expr left = stmt.left();
		Expr forinit = left.left();
		Expr forcond = left.mid();
		if(forinit && forinit.op() == T(STMTLIST)) {
			StmtList stmtlist(forinit);
			if(stmtlist.scope())
				scope(stmtlist.scope());
		} else if(forcond && forcond.op() == T(STMTLIST)) {
			StmtList stmtlist(forcond);
			if(stmtlist.scope())
				scope(stmtlist.scope());
		}
		oprintf("for");
		printexpr(stmt.left(), true);
		printstmti(Stmt(stmt.mid()));
		scope(savescope);
		} break;
	case T(KW_RETURN):
		oprintf("return ");
		printexpr(stmt.left());
		oprintf(";");
		break;
	case T(KW_GOTO):
		oprintf("goto ");
		printexpr(stmt.left());
		oprintf(";");
		break;
	case T(KW_BREAK):
		oprintf("break;");
		break;
	case T(KW_CONTINUE):
		oprintf("continue;");
		break;
	case T(KW_DEFAULT):
	case T(KW_CASE):
	case T(COLON):
		onltab(_indent-_tabwid+1);
		if(stmt.op() == T(KW_DEFAULT))
			oprintf("default:");
		else {
			if(stmt.op() == T(KW_CASE))
				oprintf("case ");
			printexpr(stmt.left());
			oprintf(":");
		}
		if(stmt.mid()) {
			onl();
			printstmt(Stmt(stmt.mid()));
		}
		break;
	case T(KW_ASM):
		oprintf("asm(");
		printexpr(stmt.left());
		if(stmt.mid() || stmt.right()) 
			oprintf(" : ");
		if(stmt.mid())
			printexpr(stmt.mid());
		if(stmt.right()) {
			oprintf(" : ");
			printexpr(stmt.right());
		}
		oprintf(");");
		break;
	case T(STMTLIST):
		printstmtlist(StmtList(stmt));
		break;
	case T(CTORINIT):
		printctorinit(StmtList(stmt.left()));
		printstmtlist(StmtList(stmt.mid()));
		break;
	default:
		printexpr(Expr(stmt));
		oprintf(";");
	}
}

void
CCOut::printstmti(Stmt stmt)
{
	if(!stmt) {
		rshift();
		onl();
		oprintf(";");
		lshift();
		return;
	}
	bool tab = (stmt.op() != T(STMTLIST));
	if(tab) {
		rshift();
		onl();
	} else
		oprintf(" ");
	printstmt(stmt);
	if(tab) 
		lshift();
}


void
CCOut::printexpr(Expr expr, bool paren = false)
{
	if(!expr) 
		return;
	if(paren)
		oprintf("(");
	bool space = true;
	sz *fmt;
	type_t t;
	switch(expr.op()) {
	case T(STMTLIST):
		printdecllist(StmtList(expr));
		break;
	case T(EXPRLIST):
		printexprlist(ExprList(expr));
		break;
	case T(KW_FOR): 
		printexpr(expr.left());
		oprintf("; ");
		printexpr(expr.mid());
		oprintf("; ");
		printexpr(expr.right());
		break;
	case T(ARRAY):
		oprintf("{");
		ws();
		printexprlist(ExprList(expr));
		ws();
		oprintf("}");
		break;
	case T(VAL_ID): {
		sz buf[MAXIDLEN];
		oprintf("%s", scoped_string(expr.decl(), buf));
		break; }
	case T(TOKENLIST):
		regurgitate(expr.tlist());
		break;
	case T(LITERAL):
		printliteral(expr.lit().type, expr.lit().val);
		break;
	case T(QMARK):
		oprintf("(");
		printexpr(expr.left(), true);
		oprintf("?");
		printexpr(expr.mid(), true);
		oprintf(":");
		printexpr(expr.right(), true);
		oprintf(")");
		break;
	case T(OP_PTRTO): case T(OP_MPTRTO):
		t = dcast_PtrTo(expr.left().rtype())->rtype();
		assert(t);
		goto printmember;
	case T(OP_DOT): case T(OP_MDOT):
		t = expr.left().rtype();
	    printmember:
		{
			while(Qualifier *qual = dcast_Qualifier(t))
				t = qual->rtype();
			Aggr *aggr = dcast_Aggr(t);
			assert(aggr);
			Expr left = expr.left();
			Token op = expr.op();
			int p = tokenprec(op, 0);
			int pleft = tokenprec(left.op(), -1);
			printexpr(left, pleft < p || !p);
			const Scope savescope = scope();
			scope(aggr->scope());
			oprintf("%s", _clex.token_string((LToken)op));
			printexpr(expr.right(), false);
			scope(savescope);
		}
		break;
	case T(OP_NEW):
		oprintf("new ");
		if(expr.right()) {
			oprintf("(");
			printexpr(expr.right());
			oprintf(") ");
		}
		{
			sz buf[MAXTYPELEN];
			buf[0] = '\0';
			type_string(expr.left().rtype(), buf);
			oprintf(strchr(buf, '(') ? "(%s)" : "%s", buf);
		}
		if(expr.mid()) {
			oprintf("(");
			printexpr(expr.mid());
			oprintf(")");
		}
		break;
	case T(OP_CAST): fmt = "(%s)"; goto printcast;
	case T(KW_CCAST): fmt = "const_cast<%s>"; goto printcast;
	case T(KW_RCAST): fmt = "reinterpret_cast<%s>"; goto printcast;
	case T(KW_DCAST): fmt = "dynamic_cast<%s>"; goto printcast;
	case T(KW_SCAST): fmt = "static_cast<%s>"; goto printcast;
	    printcast:
		{
			sz buf[MAXTYPELEN];
			buf[0] = '\0';
			type_string(expr.left().rtype(), buf);
			if(buf[0] == ':')
				strpcat(buf, " ");
			oprintf(fmt, buf);
			oprintf("(");
			printexpr(expr.right());
			oprintf(")");
		}
		break;
	case T(DECL): {
		Decl *d = expr.left().decl();
		if(d->defined())
			define_sym(d, false);
		else
			declare_sym(d, false);
		d->odeclared(true);
		d->odefined(true);
		break; }
	case T(ID):
		oprintf("%s", cstr(expr.id()));
		break;
	case T(OP_TSIZEOF):
		oprintf("sizeof(");
		printexpr(expr.right());
		oprintf(")");
		break;
	default: {
		Expr left = expr.left();
		Expr right = expr.right();
		int p = tokenprec(expr.op(), 0);
		int pleft = left ? tokenprec(left.op(), -1) : 0;
		int pright = right ? tokenprec(right.op(), 1) : 0;
		printexpr(left, pleft < p || !p);
		csz *s = 0, *e = 0;
		switch(Token op = expr.op()) {
		case T(OP_SIZEOF): s = "sizeof "; space = false; break;
		case T(OP_INCP): s = "++"; space = false; break;
		case T(OP_DECP): s = "--"; space = false; break;
		case T(OP_REF): s = "&"; space = false; break;
		case T(OP_PTR): s = "*"; space = false; break;
		case T(OP_UMINUS): s = " -"; space = false; break;
		case T(OP_UPLUS): s = " +"; space = false; break;
		case T(OP_INDEX): s = "["; e = "]"; space = false; break;
		case T(OP_COMMA): s = ", "; space = false; break;
		case T(OP_FUNC): break;
		case T(OP_ADD): case T(OP_SUB):
		case T(OP_MUL): case T(OP_DIV): case T(OP_MOD):
		case T(OP_AND): case T(OP_OR): case T(OP_XOR):
		case T(OP_NOT): case T(OP_LNOT):
		case T(OP_DOT): case T(OP_PTRTO):
		case T(OP_SCOPE): case T(OP_MDOT): case T(OP_MPTRTO):
		case T(OP_LSHIFT): case T(OP_RSHIFT):
		case T(OP_DEC): case T(OP_INC):
			space = false;
		default:
			s = _clex.token_string((LToken)op);
		}
		if(s)
			oprintf(space ?  (_x > _indent ? " %s " : "%s ")
					: "%s", s);
		printexpr(right, pright <= p || !p);
		if(e)
			oprintf("%s", e);
		}
	}
	if(paren)
		oprintf(")");
}

static Decl* dependency(Expr expr, Scope outof = 0);
static Decl* dependency(type_t type, Scope outof, bool needdef);
static Decl* dependency(Decl *decl, Scope outof, bool needdef);

static Decl*
dependency(StmtList stmtlist, Scope outof)
{
	for(Stmt s = stmtlist.first(); s; s++) {
		if(Decl *d = dependency(Expr(s), outof))
			return d;
	}
	return 0;
}

static Decl*
dependency(ExprList exprlist, Scope outof)
{
	for(Expr e = exprlist.first(); e; e++) {
		if(Decl *d = dependency(e, outof))
			return d;
	}
	return 0;
}

static Decl*
dependency(Expr expr, Scope outof)
{
	Decl *d;
	if(!expr)
		return 0;
	switch(expr.op()) {
	case T(VAL_ID): 
		d = expr.decl();
		/* XXX - this can't be enabled yet
		if(d->sinline() && !d->odefined()) {
			// XXX - need to ignore this inline for non-
			//		function calls!	
			d->oneeddef(true);
			if(!d->odeclared())
				d->oneeddecl(true);
			return d;
		}
		*/
		if(d->odeclared())
			return 0;
		if(subscope(outof, d->scope()))
			return 0;
		d->oneeddecl(true);
		return d;
	case T(TOKENLIST):
		return 0;	/* XXX */
	case T(LITERAL):
		return 0;
	case T(ARRAY):
	case T(EXPRLIST):
		return dependency(ExprList(expr), outof);
	case T(STMTLIST):
		return dependency(StmtList(expr), outof);
	case T(ID):
	case T(TYPE):
		return 0;	/* XXX */
	case T(DECL):
		return dependency(expr.left().decl(), outof, false);
	case T(OP_CAST):
	case T(KW_CCAST):
	case T(KW_DCAST):
	case T(KW_RCAST):
	case T(KW_SCAST):
	case T(OP_NEW):
		if((d = dependency(expr.left().rtype(), outof, true)))
			return d;
		return dependency(expr.right(), outof);
	case T(CTORINIT):
		if(Decl *d = dependency(expr.left(), outof))
			return d;
		if(Decl *d = dependency(expr.right(), outof))
			return d;
		return 0;
	default: 
		if(Decl *d = dependency(expr.left(), outof))
			return d;
		if(Decl *d = dependency(expr.mid(), outof))
			return d;
		if(Decl *d = dependency(expr.right(), outof))
			return d;
		return 0;
	}
}


static Decl*
dependency(type_t type, Scope outof, bool needdef)
{
	if(!type)
		return 0;
	if(DeclRef *dref = dcast_DeclRef(type)) {
		Decl *decl = dref->decl();
		while(decl->conv())
			decl = decl->conv();
		if(needdef) {
			if(!decl->odefined() && !subscope(outof, decl->scope())) {
				decl->oneeddef(true);
				return decl;
			}
		} else {
			if(!decl->odeclared() && !subscope(outof, decl->scope())) {
				decl->oneeddecl(true);
				return decl;
			}
		}
	} else if(Aggr *aggr = dcast_Aggr(type)) {
		if(Scope sc = aggr->scope()) {
			if(!outof)
				outof = sc;
			if(Decl *d = sc.get(ID_INHERIT)) {
				for(; d; d = d->link()) {
					if(!d->idecl()->odefined()) {
						d = d->idecl();
						d->oneeddef(true);
						return d;
					}
				}
			}
			Decl *pd = sc.decl();
			for(Decl *d = pd->afirst(); d; d = d->next()) {
				if(d->id() == ID_FRIEND)
					continue;
				if(d->dtype() == Decl::dt_impl)
					continue;
				if(Decl *dd = dependency(d->ptype(), outof,
						!d->declarable()))
					return dd;
			}
		}
	} else if(Array *array = dcast_Array(type)) {
		return dependency(array->rtype(), outof, needdef);
	} else if(PtrTo *ptr = dcast_PtrTo(type)) {
		return dependency(ptr->rtype(), outof, false);
	} else if(RefTo *ref = dcast_RefTo(type)) {
		return dependency(ref->rtype(), outof, false);
	} else if(Func *func = dcast_Func(type)) {
		if(Decl *d = dependency(func->rtype(), outof, needdef))
			return d;
		for(uint i=0; i<func->nparam(); i++) 
			if(Decl *d = dependency(func->ptype(i),
						outof, needdef)) 
				return d;
	} else if(Method *meth = dcast_Method(type)) {
		return dependency(meth->ftype(), outof, needdef);
	} else if(Qualifier *qual = dcast_Qualifier(type)) {
		return dependency(qual->rtype(), outof, needdef);
	}
	return 0;
}

static Decl*
dependency(Decl *decl, Scope outof, bool needdef)
{
	bool wasdefined = decl->odefined();
	bool wasdeclared = decl->odeclared();
	if(decl->storage() == Decl::sc_typedef && !decl->tagged())
		needdef = false;
	if(wasdefined)
		return 0;
	// if it is to be defined, treat it as such temporarily
	// while scanning sub-dependencies:
	decl->odefined(decl->defined());
	decl->odeclared(true);
	Decl *d;
	if(decl->dtype() == Decl::dt_impl) {
		d = dependency(decl->idecl(), outof, needdef);
		// XXX - should be more (like check all members)
	} else {
		d = dependency(decl->ptype(), outof, needdef);
		if(!d && needdef && decl->initialized()
				&& decl->storage() != Decl::sc_constant) {
			if(Func *f = dcast_Func(decl->rtype())) {
				if(!outof)
					outof = f->scope();
			}
			d = dependency(Expr(decl->vexpr()), outof);
		}
	}
	// restore it to its previous state:
	decl->odefined(wasdefined);
	decl->odeclared(wasdeclared);
	return d;
}

Decl*
CCOut::has_dependency(Decl *decl, bool needdef)
{
	if(decl->storage() == Decl::sc_friend)
		needdef = false;
	while(Decl *d = dependency(decl, 0, needdef)) {
		if(!d->nooutput(_outmask))
			return d;
		if(d->oneeddecl() && !d->odeclared())
			d->odeclared(true);
		else if(d->oneeddef() && !d->odefined() && d->defined())
			d->odefined(true);
		else {
			sz b1[MAXIDLEN];
			sz b2[MAXIDLEN];
			_err.error("%s depends on %s",
				scoped_string(decl, b1),
				scoped_string(d, b2));
			INTERNAL_ERROR;
		}
	}
	return 0;
}


void
CCOut::print_storageclass(Decl *decl)
{
	switch(decl->storage()) {
	case Decl::sc_auto:	oprintf("auto "); break;
	case Decl::sc_static:	oprintf("static "); break;
	case Decl::sc_extern:	oprintf("extern "); break;
	case Decl::sc_extern_c:	oprintf("extern \"C\" "); break;
	case Decl::sc_extern_cxx: oprintf("extern \"C++\" "); break;
	case Decl::sc_register:	oprintf("register "); break;
	case Decl::sc_typedef:	oprintf("typedef "); break;
	case Decl::sc_mutable:	oprintf("mutable "); break;
	case Decl::sc_friend:	oprintf("friend "); break;
	default: break;
	}
	if(decl->sinline()) {
		/* due to a strange gcc 2.7.2 bug, we need to make
		 * sure we don't print 'inline' when defining functions
		 * that return types that were disambiguated with the
		 * global '::' operator.  The following, for example,
		 * breaks with 2.7.2, but works with 2.8.0:
		 *	struct a {};
		 *	struct s { struct a {}; inline ::a f() const; };
		 *	inline ::a s::f() const { ::a x; return x; }
		 * removing the second 'inline' fixes the problem,
		 * so that's essentially what we're doing here.
		 */
		if(scope() == decl->scope())
			oprintf("inline "); 
		else
			oprintf("/*inline*/ ");
	}
	if(decl->sexplicit())
		oprintf("explicit "); 
	if(decl->svirtual())
		oprintf("virtual "); 
}

void
CCOut::regurgitate(TokenList *t)
{
	assert(t);

	sz c[2];
	int plev = 0;
	Item item;
	LToken tok, prevtok = (LToken)T(NONE);
	bool first = true;
	while(t->next(&item)) {
		if(first && item.token == T(COLON)) {
			rshift();
			onltab(_indent-2); oprintf(": ");
			lshift();
			continue;
		}
		first = false;
		tok = (LToken)item.token;
		switch(prevtok) {
		case T(RBRACE):
			if((tok != (LToken)T(EOS))) {
				if(tok == (LToken)T(KW_ELSE))
					oprintf(" ");
				else
					goto newline;
			}
			break;
		case T(LBRACE):
			rshift();
			goto newline;
			break;
		case T(EOS):
		newline:
			if(plev > 0)
				oprintf(" ");
			else if(tok != (LToken)T(RBRACE))
				onl();
			break;
		default:
			oprintf("%s", _clex.token_spacing(prevtok, tok));
		}
		switch(item.token) {
		case T(NUM):
			oprintf(item.c.sign ? "%lld" : "%lluU", item.val);
			for(int i=0; i<item.c.nlong; i++)
				oprintf("L");
			break;
		case T(BOOL):
			oprintf("%s", _clex.token_string(tok, (bool)item.val));
			break;
		case T(CHAR): 
			c[0] = item.val;
			c[1] = '\0';
			oprintf("%s", _clex.token_string(tok, c, 1));
			break;
		case T(STR):
		case T(ID):
			oprintf("%s", _clex.token_string(tok, cstr(item.id)));
			break;
		case T(EOS):
			oprintf(";");
			break;
		case T(LBRACE):
			if(_indent == 0)
				onl();
			oprintf("{");
			break;
		case T(RBRACE):
			lshift();
			if(prevtok == (LToken)T(LBRACE))
				oprintf("}");
			else {
				onl();
				oprintf("}");
			}
			break;
		case T(LPAREN):
			plev++;
			goto dflt;
		case T(RPAREN):
			plev--;
			goto dflt;
		default:
		dflt:
			oprintf("%s", _clex.token_string(tok));
		}
		prevtok = tok;
	}
}

void
CCOut::print_access(Decl *decl)
{
	if(decl->storage() == Decl::sc_friend)
		return;
	if(decl->dtype() == Decl::dt_impl) {
		onltab(_indent-_tabwid+1);
		oprintf("%s:", cstr(decl->id()));
		onl();
		_access = Decl::ac_public;
	}
	if(decl->access() != _access) {
		onltab(_indent-_tabwid+1);
		oprintf("%s:", decl->acname());
		onl();
		_access = decl->access();
	}
}

void
CCOut::define_aggr(Decl *decl)
{
	sz buf[MAXIDLEN];
	Aggr *aggr = (Aggr*)decl->ptype()->ptr();
	Scope scope = aggr->scope();
	if(!scope)
		return;
	Id aid = aggr->name();
	if(aid && aid != decl->id())
		oprintf(" %s", cstr(aid));
	else {
		Scope scope = aggr->scope();
		if(Decl *d = scope.get(ID_INHERIT)) {
			oprintf(" : ");
			while(d) {
				Decl *bd = d->idecl();
				if(d->access() != decl->defacc())
					oprintf("%s ", d->acname());
				oprintf("%s", scoped_string(
					bd, buf, false));
				if((d = d->link()))
					oprintf(", ");
			}
		}
		if(scope_empty(scope))
			oprintf(" {}");
		else {
			oprintf(" {");
			rshift();
			print_decls(scope, decl->defacc());
			lshift();	
			onl();
			oprintf("}");
		}
	}
}

void
CCOut::define_enum(Decl *decl)
{
	oprintf(" {");
	rshift();
	for(Decl *d = decl->afirst(); d; d = d->next()) {
		if(DeclRef *dref = dcast_DeclRef(d->ptype())) {
			Decl *d2 = dref->decl();
			if(d2 == decl) {
				onl();
				oprintf("%s = %lld,", d->cstr_id(), (int64)d->cval());
				d->odeclared(true);
				d->odefined(true);
			}
		}
	}
	lshift();
	onl();
	oprintf("}");
}

void
CCOut::print_otype(Decl *decl)
{
	switch(decl->otype()) {
	case Decl::OT_IMPLICIT:	oprintf("/*implicit*/ "); break;
	case Decl::OT_GEN:	oprintf("/*gen*/ "); break;
	default: ;
	};
}

void
CCOut::define_sym(Decl *decl, bool eos)
{
	sz buf[MAXTYPELEN+MAXIDLEN];

	if(interface()) {
		if(!decl->field() && decl->access() == Decl::ac_private)
			return;
		if(decl->storage() == Decl::sc_static)
			return;
		if(decl->isfunc()) {
			if(!decl->sinline() && !decl->purevirtual())
				return;
		} else {
			if(decl->storage() != Decl::sc_typedef
					&& decl->storage() != Decl::sc_constant)
				return;
		}
	}
	if(onlyinterface()) {
		if(decl->access() == Decl::ac_private)
			return;
		if(decl->storage() != Decl::sc_typedef 
				&& decl->storage() != Decl::sc_constant
				&& !decl->purevirtual())
			return;
	}

	if(scope().parent())
		print_access(decl);

	_curdecl = decl;

	print_otype(decl);
	if(!decl->defined()) {
		oprintf("/* ");
		declare_sym(decl);
		oprintf(" */");
		return;
	}

	if(decl->anon()) 
		strcpy(buf, decl->dtname());
	else
		strcpy(buf, scoped_string(decl, buf, true));
	const Scope savescope = scope();
	if(decl->isfunc()) {
		if(decl->storage() != TDecl::sc_typedef
		 && decl->storage() != TDecl::sc_constant
		 && _indent == 0)
			strpcat(buf, "\n");
	}
	if(decl->ptype()->istype(Type_Aggr)) {
		if(_indent == 0)
			onl(2);
		oprintf("%s", buf);
		define_aggr(decl);
	} else if(decl->dtype() == Decl::dt_enum) {
		oprintf("%s", buf);
		define_enum(decl);
	} else {
		// if(decl->scope() == savescope) 
		print_storageclass(decl);
		oprintf("%s", type_string(decl->ptype(), buf));
	}
	scope(decl->scope());
	if(decl->initialized()) {
		if(decl->isfunc() && !decl->purevirtual()) {
			Expr e = decl->vexpr();
			if(e.op() == T(STMTLIST) && StmtList(e).empty()) {
				oprintf(" {}");
				onl();
				scope(savescope);
				return;
			}
			onl();
			if(_debugexpr)
				e.dump();
			printstmt(Stmt(e));
		} else {
			if(decl->storage() == TDecl::sc_constant) {
				oprintf(" = ");
				printliteral(decl->rtype(), decl->cval());
			} else {
				if(_debugexpr)
					Expr(decl->vexpr()).dump();
				if(Expr(decl->vexpr()).op() == T(EXPRLIST)) {
					oprintf("(");
					printexpr(decl->vexpr());
					oprintf(")");
				} else {
					oprintf(decl->condasn() ? " ?= " : " = ");
					printexpr(decl->vexpr());
				}
			}
			if(eos)
				oprintf(";");
		}
	} else if(eos)
		oprintf(";");
	scope(savescope);
}


void
CCOut::declare_sym(Decl *decl, bool eos)
{
	sz buf[MAXTYPELEN+MAXIDLEN];
	csz *ss;

	_curdecl = decl;

	if(decl->dtype() == Decl::dt_impl) {
		print_access(decl);
		const Scope savescope = scope();
		if(Decl *dm = decl->afirst()) {
			scope(dm->scope());
			for(; dm; dm = dm->next()) 
				declare_sym(dm);
		}
		_access = Decl::ac_unspecified;
		scope(savescope);
		return;
	}

	if(interface()) {
		if(!decl->field() && decl->access() == Decl::ac_private)
			return;
		if(decl->storage() == Decl::sc_static)
			return;
	}
	if(onlyinterface()) {
		if(decl->access() == Decl::ac_private)
			return;
		if(decl->storage() == Decl::sc_static)
			return;
	}
	if(scope().parent())
		print_access(decl);

	print_otype(decl);
	if(!decl->isfunc() && decl->storage() != Decl::sc_typedef
			&& !decl->external() && !scope().parent()) {
		if(decl->storage() == Decl::sc_default)
			oprintf("extern ");
		else
			oprintf("/* extern */ ");
	}

	if(!decl->dtscope()) {
		Decl *d = decl->scope().decl();
		if(d && dcast_Aggr(d->ptype())) {
			if(!decl->field() && !decl->ismeth()
					&& decl->storage() != Decl::sc_friend)
				oprintf("static ");
		}
		print_storageclass(decl);
	} 
	if(decl->id() == ID_ANON) {
		if(DeclRef *dref = dcast_DeclRef(decl->ptype())) {
			if(dref->decl()->anon())
				define_sym(dref->decl(), eos);
			else {
				buf[0] = '\0';
				oprintf("%s", type_string(decl->ptype(), buf));
				if(eos) {
					oprintf(";");
					onl();
				}
			}
		} else
			INTERNAL_ERROR;
		return;
	} else if(decl->id() == ID_FRIEND) {
		ss = "";
	} else {
		// ss = cstr(decl->id());
		ss = scoped_string(decl, buf);
		while(*ss == ' ')
			++ss;
	}

	strcpy(buf, ss);
	if(decl->dtscope())
		oprintf("%s %s", decl->dtname(), buf);
	else 
		oprintf("%s", type_string(decl->ptype(), buf));
	if(decl->bitfield())
		oprintf(":%d", decl->bbsize());
	if(eos)
		oprintf(";");
	if(decl->storage() == Decl::sc_static && !decl->defined())
		oprintf("\t/* undefined */");
	if(eos)
		onl();
}


bool
CCOut::print_decls(const Scope sc, bool checkdep, bool err, int dtm)
{
	int nfail, ndec, ndef;
	bool failed = false;
	Decl *pd = sc.decl();
	Decl *lf = 0;
	do {
		nfail = ndec = ndef = 0;
		for(Decl *d = pd->afirst(); d; d = d->next()) {
			if(d->nooutput(_outmask) || d->odefined())
				continue;
			if(d->istype()) {
				if(!(dtm&DTM_TYPE))
					continue;
			} else if(d->storage() == Decl::sc_constant) {
				if(!(dtm&DTM_CONST))
					continue;
			} else {
				if(d->isfunc()) {
					if(!(dtm&DTM_FN))
						continue;
				} else {
					if(!(dtm&DTM_VAR))
						continue;
				}
				if(d->storage() == Decl::sc_static) {
					if(!(dtm&DTM_LOC))
						continue;
				} else if(d->defined()) {
					if(!(dtm&DTM_EXP))
						continue;
				} else {
					if(!(dtm&DTM_IMP))
						continue;
				}
			}
			bool dep = has_dependency(d, false);
			if(dep && checkdep && !failed) {
				if(!d->odeclared() && d->oneeddecl()
				  && !d->anon() && d->declarable()) {
					onl();
					declare_sym(d);
					d->odeclared(true);
					ndec++;
				}
				lf = d;
				++nfail;
			} else {
				onl();
				bool def = false;
				if(d->istype()) 
					def = true;
				if(d->isfunc() && d->initialized()) {
					if(d->purevirtual()) 
						def = true;
					if(d->sinline() && Expr(d->vexpr()).op() == T(STMTLIST))
						if(StmtList(d->vexpr()).empty()) 
							def = true;
				}
				if(d->storage() == Decl::sc_constant)
					def = true;
				if(!d->defined())
					def = false;

				if(def) {
					d->odeclared(true);
					++ndec;
					define_sym(d);
					d->odefined(true);
					++ndef;
				} else if(!d->anon() && !d->odeclared()) {
					declare_sym(d);
					d->odeclared(true);
					++ndec;
				}
			} 
		}
		if(nfail > 0 && ndec+ndef == 0) {
			if(!err)
				return false;
			sz b1[MAXIDLEN];
			sz b2[MAXIDLEN];
			onl();
			oprintf("/*** declaration dependency ordering failed"
						" (%s : %s) ***/",
				scoped_string(lf, b1),
				scoped_string(has_dependency(lf, false), b2));
			failed = true;
		}
	} while(nfail);
	onl();
	return true;
}



void
CCOut::print_decls(const Scope sc, TDecl::Access acc)
{
	TDecl::Access saveacc = _access;
	const Scope savescope = scope();
	_access = acc;
	scope(sc);
	bool global = !sc.parent();
	bool err;
	do {
		err = false;
		if(global)
			sectionlabel("constants:");
		if(!print_constants(sc))
			err = true;
		if(global)
			sectionlabel("types:");
		if(!print_types(sc))
			err = true;
	} while(err);
	if(global) {
		sectionlabel(hdr() ? "external:" : "imports:");
		print_fn_imports(sc);
		print_var_imports(sc);
		sectionlabel(hdr() ? "local exports/inline:" : "exports:");
		print_fn_exports(sc);	
		print_var_exports(sc);
		sectionlabel("local:");
		print_fn_local(sc);	
		sectionlabel(0);
		onl(2);
	} else {
		print_var_imports(sc);
		print_var_exports(sc);
		print_fn_imports(sc);
		print_fn_exports(sc);
	}
	_access = saveacc;
	scope(savescope);
	onl();
}



void
CCOut::print_defs(const Scope sc)
{
	int nfail, ndec, ndef;

	Decl *pd = sc.decl();

	//TDecl::Access saveacc = _access;
	//const Scope savescope = _scope;
	//_access = TDecl::ac_public;
	//_scope = sc;

	onl(2);

	do {
		nfail = ndec = ndef = 0;
		for(Decl *d = pd->afirst(); d; d = d->next()) {
			if(d->nooutput(_outmask) || !d->defined())
				continue;
			/* XXX - the !d->isfunc() should go away,
			 * but can't yet since the dependency checking
			 * runs into too many problems
			 */
			if(!d->odefined() && !d->isfunc()) {
			    if(Decl *dn = has_dependency(d, true)) {
				if(!d->odeclared() && d->oneeddecl()) {
					onl();
					declare_sym(d);
					d->odeclared(true);
					ndec++;
				}
				char buf[MAXIDLEN];
				printf("/* needs `%s' */ ", scoped_string(dn, buf));
				declare_sym(d);
				nfail++;
			    } else {
				onl();
				define_sym(d);
				d->odefined(true);
				ndef++;
				d->odeclared(true);
				ndec++;
			    }
			}
		}
		if(nfail > 0 && ndec+ndef == 0) {
			onl();
		    	oprintf("/*** definition dependency"
				 	" ordering failed ***/");
			onl();
			break;
		}
	} while(nfail);
	for(Decl *d = pd->afirst(); d; d = d->next()) {
		if(Aggr *aggr = dcast_Aggr(d->ptype())) {
			Scope scope = aggr->scope();
			if(scope && !d->nooutput(_outmask))
				print_defs(scope);
		}
	}    
	for(Decl *d = pd->afirst(); d; d = d->next()) {
		if(d->nooutput(_outmask) || !d->defined())
			continue;
		if(!d->odefined()) {
			if(d->isfunc() && !d->initialized())
				continue;
			onl(2);
			define_sym(d);
			d->odeclared(true);
			d->odefined(true);
		}
	}

	//_access = saveacc;
	//_scope = savescope;
	onl();
}


void
CCOut::print_reqs(const Scope sc)
{
	Decl *pd = sc.decl();

	sectionlabel("external declaration requirements:");
	for(Decl *d = pd->afirst(); d; d = d->next()) {
		if(!d->nooutput(_outmask))
			continue;
		if(!d->oneeddecl() || d->oneeddef())
			continue;
		oprintf("//\t");
		declare_sym(d);
		onl();
	}

	sectionlabel("external definition requirements:");
	for(Decl *d = pd->afirst(); d; d = d->next()) {
		if(!d->nooutput(_outmask))
			continue;
		if(!d->oneeddef())
			continue;
		oprintf("//\t");
		declare_sym(d);
		onl();
	}
}


CCOut::CCOut(vErrorReport &err, const Scope scope, TDecl::Access access)
	: _err(err), _clex(_buf, sizeof _buf)
{
	_scope = scope;
	_access = access;
	_strtab = scope.strtab();
	_x = 0;
	_nl = 1;
	_tabwid = 4;
	_obuf = 0;
	_fullscope = false;
	_debugtypes = false;
	_wid = 0;
	_outmask = (1<<Decl::OT_FROMHDR)|(1<<Decl::OT_HDR)|(1<<Decl::OT_CODE)
			|(1<<Decl::OT_REPOS)|(1<<Decl::OT_IMPLICIT)|(1<<Decl::OT_GEN);
	_ws = false;
}

